package ch.epfl.collections;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Hashset implementation of a generic Set.
 * 
 * The underlying implementation uses an array of single-linked Nodes. Every
 * element of the array can be considered a "bucket" which contains Nodes for the objects
 * with the same hash-value.
 * 
 * e.g. At some given moment, our array table can look like this:
 * 
 * Hash value:         0       1      2     3     4
 * 		table =     [ Node,  NULL,  Node,  NULL, Node ]
 *                     |              |
 *                    Node          Node
 *                     |
 *                    Node
 *
 * In this example, position for the hash value 0 has three nodes, each linked to the previous
 * node inserted in this bucket.
 * Hash-values 1 and 3 have no nodes and the array is empty at those positions.
 * 
 * Why are there several objects with the same hash-value? Because it's inefficient to 
 * make 4,294,967,296 buckets for all hash-values (considering that hash is 32bit integer),
 * we are limiting the size of the array, and doing the module of the hash value generated
 * by the object to fit the size of the array. This gives us higher probability of two
 * objects having the same hash-code, and thus being stored in the same bucket.
 * 
 * So, each bucket (or position in the array) is implemented as a single-linked list of nodes
 * which will allow us to store several objects with the same hash value in our set. 
 * 
 */
public class HashSet<E> implements Set<E> {
	/*
	 * We balance the length of the linked list of nodes and the size of the statically
	 * allocated array by our *_LOAD_FACTORS constants.
	 * 
	 * This gives us a balanced speed-of-access/memory-consumption. See detailed explanation
	 * in the rehashIfNeeded() method.
	 * 
	 * @see HashSet#rehashIfNeeded()
	 */
    private static final double MIN_LOAD_FACTOR = 0.5;
    private static final double IDEAL_LOAD_FACTOR = 0.75;
    private static final double MAX_LOAD_FACTOR = 1;
    private static final int MIN_CAPACITY = 10;
    
    @SuppressWarnings("unchecked")
    private Node<E>[] table = (Node<E>[])new Node[MIN_CAPACITY]; // main storage of the nodes
    private int size = 0; // the current number of elements in all buckets 
    
    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    /**
     * Adds a new element to the set.
     * 
     * The method finds the bucket based on the hash-code. If we already have a node
     * with the newElem value, we can ignore it (since this is a set - no duplicated
     * elements).
     * We are inserting the new node as the first node in the linked list of nodes,
     * and we are linking the existing nodes in that bucket to it.
     * 
     * @see HashSet#getIndex(Object)
     * @see HashSet#nodeForValue(Object, int)
     * @see HashSet#rehashIfNeeded()
     */
    @Override
    public void add(E newElem) {
        final int i = getIndex(newElem);
        final Node<E> n = nodeForValue(newElem, i);
        if (n == null) {
            table[i] = new Node<E>(table[i], newElem);
            ++size; // update the number of elements in the set
            rehashIfNeeded(); // check if we need to re-size the array
        }
    }

    /**
     * Removes an element from the set.
     * 
     * Based on the hash-code of the <code>elem</code>, the bucket is located
     * and the node containing that value is removed. The problematic part is
     * to "re-wire" other nodes in the bucket so we keep the references to all
     * nodes. This is being done by keeping track which node was before the
     * node being removed (local variable <code>prevNode</code>) and updating
     * it's <code>.next</code> element.
     * 
     * @see HashSet#getIndex(Object)
     * @see HashSet#rehashIfNeeded()
     */
    @Override
    public void remove(E elem) {
        final int i = getIndex(elem);
        Node<E> node = table[i], prevNode = null;
        // find the node containing value <code>elem</code>
        while (node != null && !node.elem.equals(elem)) {
            prevNode = node;
            node = node.next;
        }
        // if we have something to remove, re-wire the nodes to exclude
        // the node being removed
        if (node != null) {
            if (prevNode == null)
                table[i] = table[i].next; // we're removing the first node, no previous node to update
            else
                prevNode.next = node.next;
            --size; // update the number of elements in the set
            rehashIfNeeded(); // check if we need to re-size the array
        }
    }

    /**
     * Returns true if the set contains the item <code>elem</code>.
     * 
     * The method relies on the method nodeForValue which will return
     * null if the value <code>elem</code> is not found.
     * 
     * @see HashSet#nodeForValue(Object, int)
     */
    @Override
    public boolean contains(E elem) {
        return nodeForValue(elem, getIndex(elem)) != null;
    }

    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private int i = 0;
            private Node<E> n = table[i];
            private int nbLeft = size;

            @Override
            public boolean hasNext() {
                return nbLeft > 0;
            }

            @Override
            public E next() {
                if (nbLeft == 0)
                    throw new NoSuchElementException();
                while (n == null)
                    n = table[++i];
                E e = n.elem;
                n = n.next;
                --nbLeft;
                return e;
            }

            @Override
            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

    @Override
    public String toString() {
        StringBuilder b = new StringBuilder("[");
        Iterator<E> it = iterator();
        while (it.hasNext()) {
            b.append(it.next());
            if (it.hasNext())
                b.append(",");
        }
        return b.append("]").toString();
    }

    /**
     * Returns the index of the node (the "bucket") based on the hash-code
     * of the element and the length of the current array <code>table</code>.
     * 
     * Note that this means that if we change the length of the array, the
     * bin number will change, even though the hash-code remains the same.
     * For this reason we have to re-index after changing the size of the array
     * <code>table</code>.
     * 
     * @see HashSet#add(Object)
     * @see HashSet#remove(Object)
     * @see HashSet#contains(Object)
     * 
     * @return Index in the array <code>table</code> of the bucket.
     */
    private int getIndex(E elem) {
        return Math.abs(elem.hashCode()) % table.length;
    }

    /**
     * Iterates over the i-th bucket and searches for the node which contains
     * the <code>elem</code> value.
     * 
     * @param elem Value to be found
     * @param i    Index of the bucket to look into.
     * @return     The node of the element, or <code>null</code> if non is found
     */
    private Node<E> nodeForValue(E elem, int i) {
        for (Node<E> n = table[i]; n != null; n = n.next) {
            if (n.elem.equals(elem))
                return n;
        }
        return null;
    }

    /**
     * Rehashes the hash-set when the number of elements in the bins becomes too large
     * or too small.
     * 
     * Our structure dictates that we can find the bucket of any given object in constant
     * time, but to find the Node of the object itself, we have to linearly iterate over
     * the elements (nodes) in the bucket. If the bucket contains too many elements, the access
     * to the individual element will be slow, even though we can find the bucket quickly.
     * 
     * For this reason, we want to have a good ratio between the number of statically
     * allocated buckets (the size of the array <code>table</code>), and the number of nodes
     * in each bucket.
     * 
     * We manage this ratio by the <code>loadFactor</code> which we calculate as the number
     * of elements in the set divided by the number of buckets we have allocated. If the
     * <code>loadFactor</code> drops below 0.5 (meaning 50%), we are shrinking the size
     * of the HashSet to reduce the memory consumption.
     * 
     * If the <code>loadFactor</code> rises above 1.0, we are expanding the number of buckets
     * to reduce the expected time needed to access an individual element in the bucket.
     * 
     * @see HashSet#add(Object)
     * @see HashSet#remove(Object)
     */
    @SuppressWarnings("unchecked")
    private void rehashIfNeeded() {
        final double loadFactor = (double)size / (double)table.length;
        // check do we need to resize the table
        if (! ((loadFactor < MIN_LOAD_FACTOR && table.length > MIN_CAPACITY) || (loadFactor > MAX_LOAD_FACTOR)))
            return;
        
        /* new capacity is always calculated based on our IDEAL_LOAD_FACTOR, which is in this
           case 0.75  */ 
        final int newCapacity = Math.max(MIN_CAPACITY, (int) ((double)size / IDEAL_LOAD_FACTOR));
        Node<E>[] newTable = (Node<E>[])new Node[newCapacity];
        /* after the size changes, we have to re-calculate the new bucket index of every element,
           since it will change, and put elements in the new table */
        for (E elem: this) {
            final int newIndex = Math.abs(elem.hashCode()) % newCapacity;
            newTable[newIndex] = new Node<E>(newTable[newIndex], elem);
        }
        table = newTable;
    }

    /**
     * 	Internal container of values for the hash-set, implemented as a single-linked list
     *  of Nodes.
     *
     */
    private static class Node<E> {
        public Node<E> next;
        public final E elem;

        public Node(Node<E> next, E elem) {
            this.next = next;
            this.elem = elem;
        }
    }
}
